@@ -1,4 +1,4 @@
-# Copyright 2004-2021 Bright Computing Holding BV
+# Copyright 2004-2022 Bright Computing Holding BV
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,19 +14,20 @@
 
 from __future__ import absolute_import, division
 
+import concurrent.futures
 import getpass
+import itertools
 import logging
 import os
 import re
 import signal
 import socket
+import string
 import sys
 import time
 import traceback
 import yaml
 from datetime import datetime, timedelta
-from multiprocessing import TimeoutError
-from multiprocessing.pool import ThreadPool
 from subprocess import CalledProcessError, check_call
 from typing import Dict, List
 
@@ -52,6 +53,9 @@ _last_seen = 0
 
 UUID_RE = re.compile(r"^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$", re.IGNORECASE)
 
+UBUNTU_DEV_REPO = """
+deb http://dev.brightcomputing.com/bright/apt/${repo_type}/$$(ARCH)/${bcm_version}/ubuntu/${ubuntu_version}/base/ ./
+""".strip()
 
 confirm_ns = ConfigNamespace("confirm", "Confirm operations")
 confirm_ns.add_switch_parameter(
@@ -373,6 +377,8 @@ def cod_log(logger, action, percentage):
 def get_package_type(bcm_distro):
     if bcm_distro.upper().startswith("CENTOS"):
         return "yum"
+    if bcm_distro.upper().startswith("ROCKY"):
+        return "yum"
     elif bcm_distro.upper().startswith("SLES"):
         return "zypper"
     elif bcm_distro.upper().startswith("SL"):  # This has to come after SLES
@@ -443,55 +449,70 @@ def sed_file_inplace(path, old, new, plain_text=False):
         raise
 
 
-def replace_repo_version(bcm_distro, bcm_old_version, bcm_new_version, root_dir):
+def _define_dev_repository(repo, repo_type, bcm_distro, bcm_version):
+
+    major_version, _ = _extract_bcm_version(bcm_version)
+
+    if bcm_distro.lower().startswith("ubuntu"):
+
+        # The regex below is expecting it in lowercase
+        bcm_distro = bcm_distro.lower()
+        # The regex below is expecting ubuntu1604 instead of ubuntu16.04
+        bcm_distro = bcm_distro.replace(".", "")
+        # Retrieve distro/version path from distro string (e.g., ubuntu1604 -> ubuntu/1604)
+        ubuntu_version = re.match(r"([a-z]+)([0-9]+)", bcm_distro, re.I).groups()[1]
+
+        # Comment out all updates-*.brightcomputing.com repo lines
+        sed_file_inplace(
+            repo,
+            "^(deb http://([^@]+@)?updates[^.]*.brightcomputing.com)",
+            "# \\1"
+        )
+        dev_repo_line = string.Template(UBUNTU_DEV_REPO).substitute(
+            repo_type=repo_type,
+            bcm_version=major_version,
+            ubuntu_version=ubuntu_version,
+        )
+        with open(repo, "r+") as repo_file:
+            if not any([line.strip() == dev_repo_line for line in repo_file.readlines()]):
+                repo_file.write("\n" + dev_repo_line)
+    else:
+        sed_file_inplace(repo, major_version, bcm_version, plain_text=True)
+
+
+def _extract_bcm_version(bcm_version):
+    pattern = r"(.*(?=\-))(.*)"
+
     try:
-        repo_paths = list(enumerate_repo_config_files(bcm_distro, root_dir))
-        repo_paths += list(enumerate_repo_config_files(bcm_distro, root_dir, suffix="-ml", sles_suffix="_ML"))
+        return re.match(pattern, bcm_version).groups()
+    except AttributeError:
+        log.error(
+            "Failed to extract version from BCM version '%s'" % bcm_version
+        )
+        raise
 
-        for path in repo_paths:
-            if not os.path.isfile(path):
-                log.debug("Skipping '%s', does not exist" % (path))
-                continue
-            log.debug("Enabling %s repository in '%s'" % (bcm_new_version, path))
-            if bcm_distro.upper().startswith("UBUNTU"):
-                # The regex below is expecting it in lowercase
-                bcm_distro = bcm_distro.lower()
-                # The regex below is expecting ubuntu1604 instead of ubuntu16.04
-                bcm_distro = bcm_distro.replace(".", "")
-                # Retrieve distro/version path from distro string (e.g., ubuntu1604 -> ubuntu/1604)
-                distro_path = "/".join(re.match(r"([a-z]+)([0-9]+)",
-                                       bcm_distro, re.I).groups()).lower()
 
-                # The following lines replace the production apt repo with the dev apt repo
-                sed_file_inplace(path, "mirror://", "http://")
-
-                # Ideally we want to replace xenial main -> ./ in one go, but here it needs to be
-                # done in two steps because the command to subprocess.check_call() is split based
-                # on spaces in sed_file_inplace
-                sed_file_inplace(path, "focal", "")
-                sed_file_inplace(path, "bionic", "")
-                sed_file_inplace(path, "xenial", "")
-                sed_file_inplace(path, "main", "./")
-
-                # Replace default CM repos
-                old_repo_path = "deb/cm/%s/$(ARCH)/ubuntu/mirrors.txt" % (bcm_old_version)
-                new_repo_path = "bright/apt/cm/$(ARCH)/%s/%s/base" % (bcm_old_version, distro_path)
-
-                sed_file_inplace(path, old_repo_path, new_repo_path, plain_text=True)
-
-                # Replace any ML repos
-                # FIXME: this a a bit hacky. Any given file will contain either CM repos or ML repo, never both.
-                old_repo_path = "deb/ml/%s/$(ARCH)/ubuntu/mirrors.txt" % (bcm_old_version)
-                new_repo_path = "bright/apt/ml/$(ARCH)/%s/%s/base" % (bcm_old_version, distro_path)
-
-                sed_file_inplace(path, old_repo_path, new_repo_path, plain_text=True)
-            else:
-                sed_file_inplace(path, bcm_old_version, bcm_new_version, plain_text=True)
+def define_dev_repositories(bcm_distro, bcm_version, root_dir):
+    try:
+        cm_repos = enumerate_repo_config_files(bcm_distro, root_dir)
+        ml_repos = enumerate_repo_config_files(bcm_distro, root_dir, suffix="-ml", sles_suffix="_ML")
+        ni_repos = enumerate_repo_config_files(bcm_distro, root_dir, suffix="-ni", sles_suffix="_NI")
     except KeyError:
         log.warning(
-            "Not changing repository configuration files for unknown distro '%s'" % bcm_distro
+            "Not changing repository configuration files for unknown distro '%s'" % (bcm_distro, )
         )
 
+    for repo, repo_type in itertools.chain(
+        zip(cm_repos, itertools.repeat("cm")),
+        zip(ml_repos, itertools.repeat("ml")),
+        zip(ni_repos, itertools.repeat("ni"))
+    ):
+        if not os.path.isfile(repo):
+            log.debug("Skipping '%s', does not exist" % (repo, ))
+            continue
+        log.debug("Enabling %s repository in '%s'" % (bcm_version, repo))
+        _define_dev_repository(repo, repo_type, bcm_distro, bcm_version)
+
 
 def get_user_at_fqdn_hostname():
     return "{user}@{host}".format(
@@ -518,8 +539,6 @@ def multithread_run(function, args_list, max_threads):
     multithread_run(lambda f: f(), [func1, func2])
     """
 
-    pool = ThreadPool(max_threads)
-
     def wrapped(args):
         try:
             if not isinstance(args, tuple):
@@ -528,21 +547,17 @@ def multithread_run(function, args_list, max_threads):
         except Exception as e:
             return MultithreadRunError(traceback.format_exc(), e)
 
-    result = pool.map_async(wrapped, args_list)
+    with concurrent.futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
+        futures = [executor.submit(wrapped, args) for args in args_list]
+        _, not_done = concurrent.futures.wait(futures, timeout=600)
+        while not_done:
+            try:
+                _, not_done = concurrent.futures.wait(futures, timeout=600)
+            except concurrent.futures.TimeoutError:
+                log.debug("Timed out waiting for asynchronous result. Waiting again...")
+                continue
 
-    # We run result.get with a timeout, otherwise Python ignores signals (meaning that CTRL+C is ignored)
-    # Just adding a timeout is enough, so even result.get(9999999) would work, but I prefered to do the loop
-    # just in case.
-    # https://stackoverflow.com/questions/14421133/catch-keyboard-interrupt-in-program-that-is-waiting-on-an-event
-    ret = None
-    while ret is None:
-        try:
-            ret = result.get(600)
-        except TimeoutError:
-            log.debug("Timed out waiting for asynchronous result. Waiting again...")
-            continue
-
-    return ret
+    return [x.result() for x in futures]
 
 
 def running_in_virtual_env():
@@ -633,3 +648,13 @@ def report_progress(status, previous_status=""):
         sys.stderr.write("\r" + status)
 
     return status
+
+
+def parse_assignment(s):
+    """
+    :param s: a string representing an assignment, of the form key=value
+    :return: a pair of k, v. k is a non-None string, v is allowed to be empty.
+    """
+    k, v = s.split("=", 1)  # allows '=' to show up in the value
+    assert k, f"The string '{s}' is not a valid assignment in the form 'a=b'"
+    return k, v
